from pathlib import Path
from string import Template
from datetime import datetime

import re

from notion.collection import NotionDate, CollectionRowBlock
from notion.block import PageBlock

from .config import Config

"""
Functions that operate on Notion's `PageBlock`.
"""


def page_path(page: PageBlock, dest_dir: Path = Path(".")) -> Path:
    "Build a .md file path in `dest_dir` based on a Notion page metadata."
    fname = __title_to_slug(page.title) + ".md"
    return dest_dir / fname


def __title_to_slug(title: str) -> str:
    return re.sub(r"[^\w\-]", "", re.sub(r"\s+", "-", title)).lower()


def front_matter(page: CollectionRowBlock, config: Config) -> str:
    "Builds and returns a page front matter in a yaml-like format."
    # Start with all the properties from the page, including formulas and rollups.
    all_props = page.get_all_properties()

    # Add the default layout from the config
    # if there's no layout property on the page itself.
    if "layout" not in all_props:
        all_props["layout"] = config.default_layout

    # Add default published_at if there's no specific property for it.
    if "published_at" not in all_props:
        last_edited_time = datetime.utcfromtimestamp(
            int(page.get("last_edited_time")) / 1000
        )
        all_props["published_at"] = last_edited_time

    # Select only properties that are not empty
    renderables = {k: v for k, v in all_props.items() if v != ""}
    return __sanitize_front_matter(renderables)


def __sanitize_front_matter(items: dict) -> dict:
    "Sanitizes and returns front matter items as a dictionary."
    for k, v in items.items():
        if type(v) not in [str, list]:
            if isinstance(v, NotionDate):
                items[k] = v.start
            if isinstance(v, bool):
                items[k] = str(v).lower()
    return items


def page_url_substitutions(page: PageBlock, config: Config) -> dict:
    "Builds and returns a `dict` of substitutions to build URL to this page in the Blog."
    # Start with a dict from `front_m   subs = front_matter(page, config)
    subs = front_matter(page, config)

    # Grab the config baseurl
    subs["baseurl"] = config["baseurl"]

    # FIXME Clean this up into a mapping of callables?
    #
    if "title" in subs:
        subs["title"] = __title_to_slug(subs["title"])

    if "categories" in subs:
        subs["categories"] = "/".join(subs["categories"])
    else:
        subs["categories"] = ""

    if "published_at" in subs:
        subs["year"], subs["month"], subs["day"], *_ = subs["published_at"].timetuple()

    return subs


def build_page_url(page: PageBlock, pattern: Template, config: Config) -> str:
    "Build the URL string for a given URL pattern and returns it as `str`."
    return pattern.substitute(page_url_substitutions(page, config))



from pathlib import Path
from typing import Union, List

from notion.client import NotionClient
from notion.block import PageBlock
from notion.collection import Collection, NotionDate

from .config import Config
from .templates import load_template
from .page import page_path, front_matter


def notion_client(token_v2: str) -> NotionClient:
    config = Config(token_v2=token_v2)
    client = NotionClient(token_v2=config.token_v2)
    return client


def notion_blog_database(client: NotionClient, db_url: str) -> Collection:
    """
    Returns a Notion database, wraped into a `notion.Collection` for
    easy access to it's rows.
    """
    return client.get_collection_view(db_url).collection


def all_pages(blog: Collection) -> List[PageBlock]:
    "Returns all Notion PageBlocks"
    return blog.get_rows()


def published_pages(blog: Collection) -> List[PageBlock]:
    "Returns the list of published pages."
    # FIXME: This needs to be a filtered query instead.
    return [post for post in blog.get_rows() if post.get_all_properties()["published"]]


def draft_pages(blog: Collection) -> List[PageBlock]:
    "Returns the list of draft pages."
    # FIXME: This needs to be a filtered query instead.
    return [
        post for post in blog.get_rows() if not post.get_all_properties()["published"]
    ]


def page_to_markdown(page: PageBlock, config: Config) -> str:
    "Translates a Notion Page (`PageBlock`) into a Markdown string and returns it."
    return load_template("post", debug=True, config=config).render(
        page=page, front_matter=front_matter(page, config)
    )



from typing import Union
from pathlib import Path

from jinja2 import (
    Environment,
    Template,
    PackageLoader,
    select_autoescape,
    contextfilter,
)
from jinja2.runtime import Context

import notion.block as block
from notion.collection import CollectionRowBlock
from notion.markdown import notion_to_markdown

import re
import string

from .config import Config
from .page import build_page_url

"""
Provides templates that are used in Notion -> Markdown conversion.
They're currently not utilized, as the render engine actually uses
`notion.markdown.notion_to_markdown` instead
"""


def load_template(
    name: Union[str, Path], debug: bool = False, config: Config = Config()
) -> Template:
    """
    Loads the template file `templates/{name}.md.j2`.
    """

    filters = dict(
        template_name=__template_name,
        render_block=__render_block,
        snake_case=__snake_case,
        numbered_list_index=__numbered_list_index,
        notion_url=__notion_url,
        preprocess_notion_links=__preprocess_notion_links,
    )

    env = Environment(
        loader=PackageLoader("notoma", "templates"),
        autoescape=select_autoescape(["md"]),
        trim_blocks=True,
        lstrip_blocks=True,
    )

    for fl, func in filters.items():
        env.filters[fl] = func

    env.globals["debug"] = debug
    env.globals["config"] = config
    env.tests["notion_link"] = __is_notion_link
    return env.get_template(f"{name}.md.j2")


#
# Helpers
#


def __get_linked_page(this_page: block.PageBlock, page_id: str) -> CollectionRowBlock:
    "Returns the linked CollectionRowBlock by it's block id. Not exposed to the template."
    linked_page = CollectionRowBlock(this_page._client, page_id)

    if not linked_page.parent == this_page.parent:
        raise ValueError(f"The page {page_id} has different parent.")

    return linked_page


def __linked_page_url(config: Config, page: CollectionRowBlock) -> str:
    "Returns a URL of a linked page."
    pattern = string.Template(config["permalink_pattern"])

    # TODO: Add tests and remove this quite questionable behavior.
    # Or make it work on a flag, and add logging.
    if pattern is None:
        return page.get_browseable_url()

    return build_page_url(page, pattern, config)


#
# Jinja filters
#


def __template_name(block: block.Block) -> str:
    """
    Finds a Jinja template name for the provided Notion `Block`,
    checks that the template exists, and returns the template name as `str`.
    """
    template_name = f"blocks/_{__block_type(block)}.md.j2"
    template_path = Path(__file__).parent / f"templates/{template_name}"

    if template_path.exists():
        return template_name
    else:
        return None


@contextfilter
def __render_block(ctx: Context, block: block.Block) -> str:
    """
    Jinja filter that wraps the block in it's markdown equivalent if possible,
    or returns it's title.
    """
    if ctx["debug"]:
        print(f"Unsupported block type: {__block_type(block)} in {ctx['page'].title}.")
    return str(block.title)


@contextfilter
def __numbered_list_index(ctx: Context, this: block.Block) -> str:
    "Traverses the parent PageBlock and calculates the correct index for the numbered list item block."
    index = 1
    for item in ctx["page"].children:
        if isinstance(item, block.NumberedListBlock):
            if item.id == this.id:
                return index
            else:
                index += 1
        else:
            index = 1
    raise ValueError(
        f"Expected the page {ctx['page'].id} to include the provided block {this.id}"
    )


def __block_type(block: block.Block) -> str:
    "Jinja filter. Returns snake_cased block name."
    return __snake_case(block.__class__.__name__)[:-6]


def __snake_case(input_string: str) -> str:
    "Jinja filter. Returns a new string containing the input string, in camel_case."
    output = re.sub("(.)([A-Z][a-z]+)", r"\1_\2", input_string)
    return re.sub("([a-z0-9])([A-Z])", r"\1_\2", output).lower()


def __notion_url(page: block.PageBlock) -> str:
    "Jinja filter. Returns the page's (or blocks) Notion URL."
    return page.get_browseable_url()


@contextfilter
def __preprocess_notion_links(ctx: Context, b: block.TextBlock) -> str:
    config = ctx["config"]
    chunks = b.get("properties.title")

    if chunks is None or len(chunks) == 0:
        return ""

    for chunk in chunks:
        # For the chunks that have formatting
        if len(chunk) > 1:
            # Each chunk can have multiple modifiers
            for modifier in chunk[1]:
                # Only process Notion page links.
                # Replace Notion link formatting with a regular
                # Markdown link formatting and invoke Noton Py's
                # formatting function.
                if modifier[0] == "p":
                    page_id = modifier[1]
                    page = __get_linked_page(ctx["page"], page_id)
                    modifier[0] = "a"
                    modifier[1] = __linked_page_url(config, page)
                    chunk[0] = page.title

    return notion_to_markdown(chunks)


#
# Jinja Tests
#


def __is_notion_link(b: block.TextBlock) -> bool:
    "Jinja test. Returns true if the block raw title property has a Notion link in it."
    if b.title == "" or b.title is None:
        return False

    raw_title = b.get("properties.title")
    for chunk in raw_title:
        if len(chunk) > 1:
            modifier = chunk[1]
            return modifier[0][0] == "p"



from pathlib import Path
from typing import Union
import click
import nbformat
import re
from nbconvert.exporters import MarkdownExporter
from nbconvert.preprocessors import RegexRemovePreprocessor


ROOT_PATH = Path(__file__).parent.parent
NBS_PATH = ROOT_PATH / "notebooks/"
DOCS_PATH = ROOT_PATH / "docs/"

ROOT_DOCS = {"index.ipynb": "README.md", "contributing.ipynb": "CONTRIBUTING.md"}


@click.group(help="Notoma dev tools: tests and documentation generators.")
def cli():
    """
    The CLI group method wrapped in @click.group
    to invoke the dev commands.
    """
    pass


@cli.command(help="Generate documentation pages in `docs` from `notebooks`.")
def docs():
    nbs = [f for f in NBS_PATH.glob("*.ipynb")]

    for fname in nbs:
        fname = Path(fname).absolute()
        dest = Path(f"{DOCS_PATH/fname.stem}.md").absolute()
        print(f"Converting {fname} to {dest}")
        _convert_nb_to_md(fname, dest)

    for nb, dest in ROOT_DOCS.items():
        nb_path = NBS_PATH / nb
        if not nb_path.exists():
            continue
        _convert_nb_to_md(nb_path, ROOT_PATH / dest, front_matter=False)
        print(f"Converted {nb} to {dest} in repo root.")


def _get_metadata(notebook: list) -> dict:
    if not notebook["cells"]:
        raise ValueError("Expected the input to be NotebookCell-like list")

    md_cells = [c["source"] for c in notebook["cells"] if c["cell_type"] == "markdown"]
    meta = {"layout": "default"}

    for cell in md_cells:
        if cell.startswith("%METADATA%"):
            for line in cell.split("\n")[1:]:
                parts = line.split(":")
                if len(parts) > 1:
                    meta[parts[0].lower()] = parts[1].strip()
    return meta


def _convert_nb_to_md(
    fname: Union[str, Path], dest: Union[str, Path], front_matter: bool = True
) -> None:
    """
    Converts a Jupyter Notebook in `fname` to a Jekyll-compatible Markdown file
    including front matter metadata for Just The Docs.
    """
    notebook = nbformat.read(str(fname), as_version=4)
    metadata = _get_metadata(notebook)
    exporter = _build_exporter()

    prep = RegexRemovePreprocessor()
    prep.patterns = ["^%METADATA%", "^#hide"]
    notebook, _ = prep.preprocess(notebook, {})

    converted = exporter.from_notebook_node(
        notebook,
        resources={
            "meta": metadata,
            "front_matter": front_matter,
            "nb_path": f"notebooks/{Path(fname).name}",
        },
    )
    with open(str(dest), "w") as f:
        f.write(converted[0])


def _build_exporter() -> MarkdownExporter:
    """
    Build a MarkdownExporter with a custom template
    and return it.
    """
    exporter = MarkdownExporter()
    exporter.template_file = "docs-jekyll.md.j2"
    exporter.template_path.append(str(Path(__file__).parent / "templates"))
    exporter.exclude_input_prompt = False
    exporter.exclude_output_prompt = False
    exporter.exclude_unknown = False
    exporter.exclude_raw = False
    return exporter


def _make_readme(fname: Union[str, Path]):
    """
    Converts a notebook at `fname` to README.md in repository root.
    """
    _convert_nb_to_md(fname, ROOT_PATH / "README.md", front_matter=False)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['notoma']

package_data = \
{'': ['*'], 'notoma': ['templates/*', 'templates/blocks/*']}

install_requires = \
['click', 'jinja2', 'notion>=0.0.25', 'python-dotenv']

extras_require = \
{':python_version < "3.8"': ['importlib-metadata>=1.0,<2.0']}

entry_points = \
{'console_scripts': ['notoma = notoma.cli:runner',
                     'notoma-dev = notoma.dev:cli']}

setup_kwargs = {
    'name': 'notoma',
    'version': '0.2.2',
    'description': 'Write and publish your blog articles from Notion.',
    'long_description': '<!--\nTHIS FILE IS AUTOGENERATED, DON\'T EDIT.\nFile to edit instead: notebooks/index.ipynb\n-->\n\n# Notoma\n\nWrite articles for your static gen blog in Notion.\n\n\n\n<a href="https://codeclimate.com/github/xnutsive/notoma/maintainability"><img src="https://api.codeclimate.com/v1/badges/70943357e5d2c54c153a/maintainability" /></a>\n<a href="https://pypi.org/project/notoma/"><img src="https://img.shields.io/pypi/v/notoma" alt="pypi" /></a>\n![Linters](https://github.com/xnutsive/notoma/workflows/Linters/badge.svg)\n\n- [Documentation website](https://xnutsive.github.io/notoma/)\n    - [Using the CLI](https://xnutsive.github.io/notoma/using-the-cli)\n    - [Contributing](https://xnutsive.github.io/notoma/contributing)\n    - [Supported Markdown Tags](https://xnutsive.github.io/notoma/supported-markdown-tags) \n\n---\n## Install\n\nNotoma is available via Pip or Homebrew: \n\n```bash\n# Installing with pip, use this if you plan using Notoma as a python library.\npip install notoma\n```\n\nInstalling with Homebrew on Mac OS.\n\n```bash\nbrew install xnutsive/notoma/notoma\n```\n\n---\n## What can you do with Notoma\nNotoma provides commands to: \n- Convert contents of your Notion Blog database to a bunch of Markdown files.\n- *Coming soon*: Watch Notion Blog database for updates and regenerate Markdown files on any updates.\n- *Coming soon*: Create a new Notion database for your Blog with all required fields.\n\nBasic usage example: this command will convert only published posts from a Notion blog database to the `./posts/ directory`.\n\n```bash\nnotoma convert --dest ./posts/\n```\n\nThis example assumes that you have a `.env` config file with authentication and blog url parameters in it.\n\n#### Authenticating in Notion\n\nNotoma uses an internal Notion API, and that, unfortunately, requires you to provide an authentication token `token_v2` that you can find in your notion.so cookes.\n\nYou can provide `token_v2` option to every command line call, or store it in your environment, or [`.env` config file](.env.sample).\n\n---\n## Notion database structure\nNotoma has very few expectations about how your Notion is structured. Here\'s a [public example database](https://www.notion.so/respawn/7b46cea379bd4d45b68860c2fa35a2d4?v=b4609f6aae0d4fc1adc65a73f72d0e21).\n\nNotoma requires that your Notion blog database has the following **properties**:\n- **Published**: whether the article is published, or is still a draft\n- **Title**: Will be used to create a file name for that article\'s Markdown equivalent file. *Won\'t be used in the article itself.*\n\nNotoma tries to parse other properties and add them as front matter into the resulting Markdown articles: \n- **Published at** will be used as publicataion date for the article, if present.\n- **Categories** will be used as `categories` front matter key, so it\'s expected to be a **multiple choice** propery.\n\n',
    'author': 'Nate Gadzhibalaev',
    'author_email': 'nate@respawn.io',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://xnutsive.github.io/notoma',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'entry_points': entry_points,
    'python_requires': '>=3.6',
}


setup(**setup_kwargs)
